/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2000 France Telecom R&D
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Bruno Dumant
 * 
 * -----------------------------------------------------------------
 * $Id: NameHelpers.java,v 1.1.1.1 2002/08/06 13:13:48 bmi Exp $
 * -----------------------------------------------------------------
 */

package org.objectweb.jonathan.libs.helpers;

import org.objectweb.jonathan.apis.kernel.Name;

/** 
 * Somme helper functions on {@link Name names}.
 */
public final class NameHelpers {


   /** 
    * Returns the suffix of <code>_name</code>.
    * <p>
    * If <code>_name</code> may be represented as "a/b/c", with '/' as a separator,
    * then getSuffix returns "c".
    * @param _name a name;
    * @return      the last component of name.
    */
   public static Name getSuffix(Name _name){
      Name $current = _name, $previous = null;
      while($current != null){
         $previous = $current;
         $current = $current.getSubName();
      }
      return $previous;
   }
   
   /** 
    * Returns the prefix of <code>_name</code>.
    * <p>
    * If <code>_name</code> may be represented as "a/b/c", with '/' as a separator,
    * then getPrefix returns "a/b".
    * @param _name a name;
    * @return      the prefix composed of all the name components but the last one.
    */
   public static Name getPrefix(Name _name){
      Name $subname = _name.getSubName();
      if ($subname == null) {
         return null;
      } else {
         return new NameImpl(_name.getId(),getPrefix($subname));
      }
   }

   /** 
    * Appends _name1 and _name2.
    * @param _name1 a name;
    * @param _name2 another name;
    * @return the concatenation of both names.
    */
   public static Name append(Name _name1, Name _name2) {
      Name $subname = _name1.getSubName();
      if ($subname == null) {
         return new NameImpl(_name1.getId(),_name2);
      } else {
         return new NameImpl(_name1.getId(),append($subname,_name2));
      }
   }

   public static Name append(Name _name1, String _id, char _separator){
      return append(_name1,newName(_id,_separator));
   }

   public static String toString(Name _name, char _separator){
      if (_name != null){
         Name $subName = _name.getSubName();
         if ($subName==null) return _name.getId();
         return _name.getId()+_separator+toString($subName,_separator);
      } else {
         return "null";
      }
   }

   /** 
    * Returns a new {@link Name Name}, using <code>_id</code> as identifier, 
    * and <code>_subname</code> as subname.
    * @param _id      a String identifier;
    * @param _subname a name;
    * @return         a new {@link Name Name}, using <code>_id</code> as identifier,
    *                 and <code>_subname</code> as subname.
    */
   public static Name newName(String _id, Name _subname){
      return new NameImpl(_id,_subname);
   }
      
   
   /** 
    * Creates a new name by parsing the provided string.
    * <p>
    * If <code>separator == 0</code>, <code>_name</code> is considered as a 
    * simple identifier, and this call returns {@link #newName(String,Name) 
    * <code>newName(_name,null)</code>}.
    * If the name begins with the separator, the name is an absolute name (this 
    * only makes sense if contexts with a tree structure); otherwise, the name is
    * relative.
    *
    * @param _name     a string representing a name;
    * @param separator a separator to parse it.
    * @return          a {@link Name Name}.
    */
   public static Name newName(String _name,char separator){
      if (_name.length() > 0 && _name.charAt(0) == separator) {
         return new NameImpl("",newRelativeName(_name,1,separator));
      } else if (_name.length() == 0) {
         return new NameImpl(_name,null);
      } else {
         return newRelativeName(_name,0,separator);
      }
   }
   
   static Name newRelativeName(String _name, int _offset,char _separator) {
      if (_offset >= _name.length()) {
         return null;
      }
      int $index = _name.indexOf(_separator,_offset);
      if ($index == -1){
         return new NameImpl(_name.substring(_offset),null);
      } else {
         return new NameImpl(_name.substring(_offset,$index),
                             newRelativeName(_name,$index+1,_separator));
      }
   }

   static class NameImpl implements Name {
      String id;
      Name subname;
      
      NameImpl(String _id, Name _subname) {
         id = _id;
         subname = _subname;
      }

      public String getId() {
         return id;
      }

      public Name getSubName() {
         return subname;
      }

      public boolean equals(Object _obj) {
         if (_obj instanceof Name) {
            Name $obj = (Name) _obj;
            if (id.equals($obj.getId())) {
               if (subname == null) {
                  return $obj.getSubName() == null;
               } else {
                  return subname.equals($obj.getSubName());
               }
            }
         } 
         return false;
      }

      public int hashCode() {
         return id.hashCode() + (subname == null ? 0 : subname.hashCode());
      }

      public String toString() {
         if (subname != null) {
            return new StringBuffer(id).append('/').append(subname.toString()).
                   toString();
         } else if (id.equals("")) {
            return "/";
         } else {
            return id;
         }
      }
   }
}
