/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2000 France Telecom R&D
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Release: 3.0
 *
 * Contact: jonathan@objectweb.org
 *
 * Author: Bruno Dumant
 * 
 */

package org.objectweb.jonathan.libs.kernel;

import org.objectweb.jonathan.apis.kernel.JonathanException;
import org.objectweb.jonathan.apis.kernel.InternalException;
import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.Factory;

import org.objectweb.jonathan.libs.helpers.HTable;

/** 
 * Generic {@link Factory Factory} implementation.
 * <p>
 * This class provides a generic mechanism to use a pool of objects and let 
 * already constructed objects be reused.
 */
public abstract class GenericFactory implements Factory {

   static HTable factories = HTable.newTable();

   /** 
    * Returns an object created using information contained in the provided 
    * {@link Context context}.
    *
    * This method first calls the {@link #getUsedComponents(Context) 
    * getUsedComponents()} method to retrieve the components that would be used 
    * to create a new object. If an object has already been constructed using
    * the same components, it is returned. Otherwise, the 
    * {@link #newInstance(Context,Object[]) newInstance} method is called to create
    * a new object. Before being returned, the object is stored, so that it can be 
    * reused in future calls.
    * 
    * @param _c a {@link Context Context} object;
    * @return an object created using information contained in <code>context</code>.
    * @exception JonathanException if something goes wrong.
    */
   public Object newObject(Context _c) throws JonathanException {
      synchronized (GenericFactory.class) {
         Class $factory_class = getClass();
         Holder $existing = (Holder) factories.get($factory_class);

         Holder $holder = $existing;
         while ($holder != null) {
            if ($holder.configuration == _c) {
               if ($holder.instance == Context.NO_VALUE) {
                  // recursive call
                  $holder.instance = newInstance();  
               }
               return $holder.instance;
            } else {
               $holder = $holder.next;
            }
         }
         
         $holder = new Holder(_c,$existing);
         factories.put($factory_class,$holder);

         Object[] $found_components = getUsedComponents(_c);

         Object $instance = Context.NO_VALUE;
         if ($holder.instance == Context.NO_VALUE) {
            $instance = newInstance(_c,$found_components); 
            $holder.instance = $instance;
         } else {
            // recursive case
            $instance = $holder.instance;
            configure($instance,_c,$found_components);
         }
         return $instance;
      }
   }

   /** 
    * Registers an instance, using the provided components, in the factory.
    * @param object the instance of an object that could have been created by
    *        the target factory;
    * @param used_components the components is used to create the object.
    */
   public void register(Object object, Context _configuration) {
      synchronized (GenericFactory.class) {
         Class $factory_class = getClass();
         Holder $holder = (Holder) factories.get(getClass());
         factories.put($factory_class,new Holder(_configuration,object,$holder));
      }
   }

   /** 
    * Returns the components of the provided context that should be used to build 
    * a new object instance.
    * 
    * @param _c a context;
    * @return the components of <code>_c</code> that should be used to build a new
    *         object 
    */
   abstract protected Object[] getUsedComponents(Context _c);


   /** 
    * Returns a new object created using the provided components, and possibly 
    * some information contained in the provided context.
    * <p>
    * The provided components must have been obtained using the 
    * {@link #getUsedComponents(Context) getUsedComponents} method.
    *
    * @param _c a context; 
    * @param _used_components components to be used to create a new object;
    * @return a new object created using <code>_used_components</code>; 
    * @exception JonathanException if something goes wrong.
    */
   abstract protected Object newInstance(Context _c,Object[] _used_components)
      throws JonathanException ;

   protected Object newInstance() throws JonathanException {
      throw new InternalException("Recursive configurations are not handled.");
   }  

   protected void configure(Object _object,Context _c, Object[] _used_components) 
      throws JonathanException {
      throw new InternalException("Recursive configurations are not handled.");
   }

   final static class Holder {
      Context configuration;
      Object instance;
      Holder next;
      
      Holder(Context _configuration,Holder _next) {
         configuration = _configuration;
         instance = Context.NO_VALUE;
         next = _next;
      }
      Holder(Context _configuration,Object _instance,Holder _next) {
         configuration = _configuration;
         instance = _instance;
         next = _next;
      }
   }
}

