/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2000 France Telecom R&D
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Fabien Delpiano, Bruno Dumant
 * 
 * --------------------------------------------------------------------------
 * $Id: JConfigurationFactory.java,v 1.1.1.1 2002/08/06 13:13:49 bmi Exp $
 * -------------------------------------------------------------------------- 
 */

package org.objectweb.jonathan.libs.kernel;

import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.Component;
import org.objectweb.jonathan.apis.kernel.ConfigurationFactory;
import org.objectweb.jonathan.apis.kernel.Element;
import org.objectweb.jonathan.apis.kernel.Name;
import org.objectweb.jonathan.apis.kernel.Factory;
import org.objectweb.jonathan.apis.kernel.ClassPathFinder;
import org.objectweb.jonathan.apis.kernel.JonathanException;
import org.objectweb.jonathan.apis.kernel.InternalException;

import org.objectweb.jonathan.libs.helpers.NameHelpers;

import java.util.Properties;
import java.util.Enumeration;



/** 
 * Configuration Factory used by {@link KKernel KKernel}.
 */
public class JConfigurationFactory extends TreeContextFactory implements ConfigurationFactory {

   static JConfiguration[] pool = new JConfiguration[10];
   static int pool_size = 0;   

   public static final JConfigurationFactory instance = new JConfigurationFactory();
   
   public Context newContext() {
      synchronized (JConfigurationFactory.class) {
         if (pool_size > 0) {
            JConfiguration configuration = pool[--pool_size];
            configuration.acquire();
            return configuration;
         } else {
            return new JConfiguration();
         }
      }
   }

   public Context newContext(Context _scope) {
      synchronized (JConfigurationFactory.class) {
         if (pool_size > 0) {
            JConfiguration configuration =  pool[--pool_size];
            configuration.acquire();
            configuration.enclosing_context = _scope;
            return configuration;
         } else {
            return new JConfiguration(_scope);
         }
      }
   }

   public Context newContext(ClassPathFinder class_finder) {
      synchronized (JConfigurationFactory.class) {
         if (pool_size > 0) {
            JConfiguration configuration =  pool[--pool_size];
            configuration.acquire();
            configuration.setClassPathFinder(class_finder);
            return configuration;
         } else {
            return new JConfiguration(class_finder);
         }
      }
   }

   public Context newContext(Context scope,ClassPathFinder class_finder) {
      synchronized (JConfigurationFactory.class) {
         if (pool_size > 0) {
            JConfiguration configuration =  pool[--pool_size];
            configuration.acquire();
            configuration.enclosing_context = scope;
            configuration.setClassPathFinder(class_finder);
            return configuration;
         } else {
            return new JConfiguration(scope,class_finder);
         }
      }
   }

   protected class JConfiguration extends TreeContext {

      protected JConfiguration(){
         super();
      }

      protected JConfiguration(Context _scope){
         super(_scope);
      }

      protected JConfiguration(ClassPathFinder _finder){
         super();
         setClassPathFinder(_finder);
      }
      
      protected JConfiguration(Context _scope,ClassPathFinder _finder){
         super(_scope);
         setClassPathFinder(_finder);
      }
      
      void setClassPathFinder(ClassPathFinder _class_finder) {
         if (_class_finder != null) {
            try {
               addElement("classpath_finder",ClassPathFinder.class,_class_finder,
                          (char) 0);
            } catch (JonathanException e) {
               throw new InternalException(e);
            }
         }
      }
      
      public String toString() {
         return "JConfiguration[" + id + "," + getAbsoluteName() + 
            (enclosing_context != null ? "," + enclosing_context + "]]" : "]");
      }
      
      public void reset() {
         super.reset();
         unused = null;
      }

      protected void recycle() {
         synchronized (JConfigurationFactory.class) {
            if (pool_size < pool.length) {
               pool[pool_size++] = this;
            }
         }
      }

      protected Element getElement(String name,boolean local) {
         Element element = super.getElement(name,local);
         if (local || name.length() == 0) {
            return element;
         } else {
            return createLocalElement(element);
         }
      }
      
      Element createLocalElement(Element element) {
         if (element == null) {
            return null;
         }
         String $name = element.getName();
         if ($name.equals(".") || $name.equals("..") || $name.equals("")) {
            return element;
         }
         if (element.getContainer() != this) {
            Component $component = element.getComponent();
            Component $component_copy = $component instanceof Forkable ?
               ((Forkable)$component).internalFork() : 
               (($component != null) ? $component.fork() : null);
            if ($component != $component_copy) {
               ContextElement $new_element = newContextElement(element.getName(), $component_copy, this);
               try {
                  addLocalElement($new_element,false);
               } catch (JonathanException e) {
                  throw new InternalException(e);
               }
               return $new_element;
            } else {
               return element;
            }
         } else {
            //Same context : return found element
            return element;
         }
      }

      protected Element handleScopeElement(Element element) {
         return createLocalElement(element);
      }
   }

}


