/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Bruno Dumant
 * 
 * --------------------------------------------------------------------------
 * $Id: OpenXMLDecoder.java,v 1.1.1.1 2002/08/06 13:13:53 bmi Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonathan.libs.kernel.decoders;

import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.ContextFactory;
import org.objectweb.jonathan.apis.kernel.Element;
import org.objectweb.jonathan.apis.kernel.Component;
import org.objectweb.jonathan.apis.kernel.ClassPathFinder;
import org.objectweb.jonathan.apis.kernel.JonathanException;

import org.objectweb.jonathan.apis.kernel.decoders.File2Configuration;

import org.objectweb.jonathan.libs.helpers.HTable;
import org.objectweb.jonathan.libs.kernel.JComponent;

import nanoxml.XMLElement;
import nanoxml.XMLParseException;

import java.util.Properties;
import java.util.Enumeration;
import java.util.List;
import java.util.LinkedList;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;


/**
 * XML to runtime components translator. Has built-in alias, atom,
 * assemblage, include & property decoders. May handle any other
 * component types as long as their specific decoders are provided.
 * <P>
 * In order to be able to decode a given custom-made component type,
 * a specific TagCompiler must be implemented for this type, and
 * provided in the <code>OpenXMLDecoder</code> configuration context
 * (i.e. in its constructor Context parameter) as the value of
 * an Element named after the component type's XML tag.
 * <P>
 * An instance of <code>OpenXMLDecoder</code> is automatically
 * used at startup by Jonathan's {@link org.objectweb.jonathan.libs.kernel.KKernel} to decode
 * Jonathan's runtime configuration file. A complete configuration
 * of <code>OpenXMLDecoder</code> is therefore provided in
 * Jonathan's compiled boot configuration.
 * @see org.objectweb.jonathan.libs.kernel.KKernel
 */
public class OpenXMLDecoder implements File2Configuration {

   /** the built-in decoders */
   static HTable default_decoders = HTable.newTable();
   
   static {
      default_decoders.put("ALIAS",new XMLAliasDecoder());
      default_decoders.put("ATOM",new XMLAtomDecoder());
      default_decoders.put("ASSEMBLAGE",new XMLAssemblageDecoder());
      default_decoders.put("INCLUDE",new XMLIncludeDecoder());
      default_decoders.put("PROPERTY",new XMLPropertyDecoder());
   }
   
   /** the additional specific decoders */
   Context decoders;

   /**
    * creates a new xml decoder with the given additional specific decoders
    * @param _decoders contains the additional specific decoders; their elements
    * have the name of their component types' XML tags.
    */
   public OpenXMLDecoder(Context _decoders) {
      decoders = _decoders;
   }
   
   public OpenXMLDecoder() {
      decoders = null;
   }
   
   /**
    * imports an XML file in the given configuration. First reads the root
    * XML element, and then dispatches its decoding to a new
    * Dispatcher.
    * @param _configuration the destination configuration context
    * @param _xml_file_name the XML source file name
    * @param _context_factory used to create new {@link Context}s
    * @param _finder used to find the XML file (that's all??)
    */
   public void importElement(Context _configuration,
                             String _xml_file_name,
                             ContextFactory _context_factory,
                             ClassPathFinder _finder)
      throws JonathanException {
      XMLElement $xml_element = readXMLElementFromResource(_xml_file_name,_finder);
      String $current_tag = $xml_element.getName();
      Dispatcher $dispatcher = new Dispatcher(_context_factory,decoders,_finder);
      if ($current_tag.equals("CONFIGURATION")) {
         $dispatcher.importXMLElements($xml_element.enumerateChildren(),_configuration);
      } else {
         $dispatcher.importXMLElement($xml_element,_configuration);
      }
   }

   /**
    * Reads a resource and parses it as an XML file. Used by
    * <code>importElement</code>.
    * @param _xml_file_name the XML source file name
    * @param _finder used to find the XML files to be parsed.
    * @return the corresponding NanoXML element.
    */
   XMLElement readXMLElementFromResource(String _resource_name,ClassPathFinder _finder) 
      throws JonathanException {
      try {
         InputStream $is = _finder.findResource(_resource_name);
         if ($is == null) {
            throw new JonathanException("Resource " + _resource_name + " not found.");
         }
         XMLElement $result = new XMLElement();
         $result.parseFromReader(new BufferedReader(new InputStreamReader($is)));
         return $result;
      } catch (IOException e) {
         throw new JonathanException(e);
      } catch (XMLParseException e) {
         throw new JonathanException(e);
      }
   }


   /**
    * Dispatches XML elements decoding among default & specified runtime
    * component decoders. Used internally by {@link OpenXMLDecoder}.
    */
   static class Dispatcher implements XMLElementDecoderDispatcher {
      
      ContextFactory context_factory;
      Context specified_decoders;
      HTable used_decoders;
      ClassPathFinder finder;

      /**
       * Creates a new decoding dispatcher.
       * @param _context_factory used to create new {@link Configuration}
       * components
       * @param _specified_decoders initially specified XML to runtime
       * component decoders
       * @param _finder finds the included XML files.
       */
      Dispatcher(ContextFactory _context_factory,Context _specified_decoders,ClassPathFinder _finder) {
         context_factory = _context_factory;
         specified_decoders = _specified_decoders;
         finder = _finder;
         used_decoders = HTable.newTable();
      }
      
      /**
       * Converts a XML element to its corresponding runtime component.
       * @param _xml_element to be decoded
       * @param _element_name the component's element name
       * @param _configuration where the component will be created
       * @param return the corresponding decoded runtime component
       */
      public Component decode(XMLElement _xml_element,String _element_name,Context _configuration) 
         throws JonathanException {
         String $current_tag = _xml_element.getName();
         if ($current_tag.equals("CONFIGURATION")){
            Context $new_configuration = null;
            if (_element_name != null) {
               $new_configuration = _configuration.addOrGetContext(_element_name,'/');
            } else {
               $new_configuration = context_factory.newContext();
            }
            
            String $scope_name = _xml_element.getStringAttribute("SCOPE");
            if ($scope_name != null) {
               Context $scope = _configuration.addOrGetContext($scope_name,'/');
               $new_configuration.setScope($scope);
            }
            importXMLElements(_xml_element.enumerateChildren(),$new_configuration);
            return $new_configuration;
         } else {
            return getDecoder($current_tag).decode(_xml_element,_element_name,_configuration,this);
         }
      }
      
      public Component error(String message) throws JonathanException {
         throw new JonathanException(message);
      }

      /**
       * returns the dispatcher's class finder.
       * Used by the {@link XMLIncludeDecoder} to find the included XML files.
       */
      public ClassPathFinder getFinder() {
         return finder;
      }
      
      /** 
       * returns the dispatcher's context factory.
       * Used by the {@link XMLIncludeDecoder} to create the included
       * properties files.
       */
      public ContextFactory getContextFactory(){
         return context_factory;
      }
      
      /**
       * returns the corresponding XML element to component decoder.
       * First tries to find a specified decoder; if this search fails,
       * tries to get a corresponding built-in decoder.
       * @param _tag the XML tag whose decoder has to be found
       * @return the found specified or built-in decoder
       * @throws JonathanException if no decoder was found
       */
      XMLElementDecoder getDecoder(String _tag) throws JonathanException {
         Object $decoder = used_decoders.get(_tag);
         if ($decoder == null) {
            $decoder = specified_decoders != null ? specified_decoders.getValue(_tag,(char) 0) : null;
            if (! ($decoder instanceof XMLElementDecoder)) {
               $decoder = default_decoders.get(_tag);
               if ($decoder == null) {
                  throw new JonathanException("No decoder found for tag: " + _tag + ".");
               }
            }
            used_decoders.put(_tag,$decoder);
         }
         return (XMLElementDecoder) $decoder;
      }
      

      /**
       * imports runtime components from XML elements in the given
       * configuration.
       * @throws JonathanException if something gets wrong.
       */
      void importXMLElements(Enumeration _xml_elements, Context _configuration)
         throws JonathanException { 
         while (_xml_elements.hasMoreElements()) {
            XMLElement $xml_element = (XMLElement) _xml_elements.nextElement();
            importXMLElement($xml_element,_configuration);
         }
      }

      /**
       * imports a single runtime component from its XML elements in the given
       * configuration.
       * @throws JonathanException if something gets wrong.
       */
      void importXMLElement(XMLElement _xml_element, Context _configuration)
         throws JonathanException  {
         String $current_tag = _xml_element.getName();
         
         if ($current_tag.equals("DOCUMENTATION")){
            return;
         }
         if (! $current_tag.equals("ELEM")) {
            error("All elements must be defined with tag \"ELEM\".");
         }
         if (_xml_element.countChildren() == 0) { //unset component
            //skip
            return;
         }
         String $element_name = _xml_element.getStringAttribute("NAME");
         
         XMLElement $xml_element = (XMLElement) _xml_element.getChildren().elementAt(0);
         Component $component = decode($xml_element,$element_name,_configuration);
         _configuration.addElement($element_name,$component,'/');
      }
   }   
}
