/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Bruno Dumant
 *
 * -------------------------------------------------------------------
 * $ID$
 * -------------------------------------------------------------------
 * 
 */

package org.objectweb.jonathan.libs.kernel.decoders;

import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.Component;
import org.objectweb.jonathan.apis.kernel.JonathanException;

import nanoxml.XMLElement;

import org.objectweb.jonathan.libs.kernel.JComponent;

/**
 * Property specific XML to runtime component decoder.
 * Used by the {@link OpenXMLDecoder} as one of the built-in
 * component decoders provided it provides to its decoding
 * dispatcher.
 */
public class XMLPropertyDecoder implements XMLElementDecoder {
   
   /**
    * Converts a XML <code>PROPERTY</code> element to its
    * corresponding property runtime component.<P>
    * Supported property types are:
    * <UL><LI>String</LI>
    * <LI>char</LI>
    * <LI>boolean</LI>
    * <LI>byte</LI>
    * <LI>short</LI>
    * <LI>int</LI>
    * <LI>long</LI>
    * <LI>float</LI>
    * <LI>double</LI></UL>
    * @param _xml_element <code>PROPERTY</code> tag to be decoded
    * @param _element_name the property's element name
    * @param _configuration where the property will be created
    * @param return the corresponding decoded runtime property
    * component
    */
   public Component decode(XMLElement _xml_element, String _element_name, Context _configuration,
                           XMLElementDecoderDispatcher _dispatcher) 
      throws JonathanException {
      
      String $current_tag = _xml_element.getName();

      if ($current_tag.equals("PROPERTY")){
         String $element_type = _xml_element.getStringAttribute("TYPE");
         String $element_value = _xml_element.getStringAttribute("VALUE");

         if ($element_type.equals("String")){
            return new JComponent(String.class,$element_value);

         } else if ($element_type.equals("char") ||
                    $element_type.equals("Character")) {
            if ($element_value==null || $element_value.length()!=1) {
               return _dispatcher.error("Value " + $element_value + " on line " + _xml_element.getLineNr() + 
                                        " must be parsable as a character.");
            }
            return new JComponent(Character.class,
                                  new Character($element_value.charAt(0)));

         } else if ($element_type.equals("boolean") ||
                    $element_type.equals("Boolean")){
            if ($element_value.equalsIgnoreCase("true")) {
               return new JComponent(Boolean.class,Boolean.TRUE);
            } else if ($element_value.equalsIgnoreCase("false")) {
               return new JComponent(Boolean.class,Boolean.FALSE);
            } else {
               return _dispatcher.error("Value " + $element_value + 
                                        " on line " + _xml_element.getLineNr() + 
                                        " must be either 'true' or 'false'.");
            }

         } else if ($element_type.equals("byte") ||
                    $element_type.equals("Byte")){
            try {
               Byte $b = new Byte($element_value);
               return new JComponent(Byte.class,$b);
            } catch (NumberFormatException e) {
               return _dispatcher.error("Value " + $element_value + " on line " + _xml_element.getLineNr() + 
                                        " must be parsable as a byte.");
            }

         } else if ($element_type.equals("short") ||
                    $element_type.equals("Short")){
            try {
               Short $s = new Short($element_value);
               return new JComponent(Short.class,$s);
            } catch (NumberFormatException e) {
               return _dispatcher.error("Value " + $element_value + " on line " + _xml_element.getLineNr() + 
                                        " must be parsable as a short.");
            }

         } else if ($element_type.equals("int") ||
                    $element_type.equals("Integer")){
            try {
               int $i = Integer.parseInt($element_value);
               return new JComponent(int.class,$i);
            } catch (NumberFormatException e) {
               return _dispatcher.error("Value " + $element_value + " on line " + _xml_element.getLineNr() + 
                                        " must be parsable as an integer.");
            }

         } else if ($element_type.equals("long") ||
                    $element_type.equals("Long")){
            try {
               Long $l = new Long($element_value);
               return new JComponent(Long.class,$l);
            } catch (NumberFormatException e) {
               return _dispatcher.error("Value " + $element_value + " on line " + _xml_element.getLineNr() + 
                                        " must be parsable as a long.");
            }

         } else if ($element_type.equals("float") ||
                    $element_type.equals("Float")){
            try {
               Float $f = new Float($element_value);
               return new JComponent(Float.class,$f);
            } catch (NumberFormatException e) {
               return _dispatcher.error("Value " + $element_value + " on line " + _xml_element.getLineNr() + 
                                        " must be parsable as a float.");
            }

         } else if ($element_type.equals("double") ||
                    $element_type.equals("Double")){
            try {
               Double $d = new Double($element_value);
               return new JComponent(Double.class,$d);
            } catch (NumberFormatException e) {
               return _dispatcher.error("Value " + $element_value + " on line " + _xml_element.getLineNr() + 
                                        " must be parsable as a double.");
            }

         } else if ($element_type.equals("null")) {
            return new JComponent(Object.class,null);

         } else {
            return _dispatcher.error("Unsupported property type! See at line "
                                     +_xml_element.getLineNr()+ " where type " +
                                     $element_type +" is used.");
         }

      } else {
         return _dispatcher.error("XMLPropertyDecoder error: Wrong tag " + $current_tag);
      }
   }
}
