/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2000 France Telecom R&D
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Fabien Delpiano
 *
 * with contributions from :
 *      Bruno Dumant, Marc Dutoo
 * 
 */

package org.objectweb.jonathan.tools;

import java.util.*;
import java.io.*;

import org.objectweb.jonathan.apis.kernel.JonathanException;
import org.objectweb.jonathan.tools.compilers.TagCompiler;
import nanoxml.XMLElement;
import nanoxml.XMLParseException;

/** 
 * Kcf2java (formerly called XML2Kernel) is a utility that compiles an XML configuration file into a java 
 * class, and possibly compiles it and adds the compiled file to a jar archive. 
 * This utility expects that the <code>javac</code> and <code>jar</code> commands 
 * are available locally.
 * <p>
 * The name of the java class to generate should be specified as the "generated"
 * attribute of the configuration element; the class used to create the actual 
 * bootstrap configuration should be specified using the "bootstrap" attribute, 
 * like in :
 * <quote>
 * <CONFIGURATION bootstrap="org.objectweb.jonathan.libs.kernel.KKernel"
 *                generated="org.objectweb.jonathan.apis.kernel.Kernel">
 *    ...
 * </CONFIGURATION>
 * </quote>
 * <p>
 * Usage: <code>java org.objectweb.jonathan.tools.Kcf2java [options] configuration-file</code>
 * where options are:
 * <ul>
 * <li> <code>-noc</code> :
 *
 *      if the generated java file should not be compiled.
 * <li> <code>-s srcdir</code>:
 *
 *      <code>srcdir</code> is the (existing) directory where
 *      the java  file should be generated;
 * <li> <code>-d classdir</code>:
 *
 *      If classdir ends with ".jar", the java file is compiled locally (with
 *      option <code>-d .</code>) and the generated file is added to the mentioned 
 *      jar file; otherwise,  this option is passed to javac when compilation 
 *      occurs. 
 * <li> <code>-p properties_file</code>:
 *
 *       where Kcf2java finds which custom XML tag compilers it should
 *       also use.
 * <li> <code>-verbose</code>:
 *
 *       tells Kcf2java to output messages on stderr.
 *       
 * </ul>
 * <P>
 * Kcf2java has already a built-in configuration decoder, and already
 * knows the following default tag compilers:
 * <UL>
 *   <li>the XML <B>ATOM</B> {@link
 * org.objectweb.jonathan.tools.compilers.AtomCompiler tag compiler}
 *   <li>the XML <B>ALIAS</B> {@link
 * org.objectweb.jonathan.tools.compilers.AliasCompiler tag compiler}
 *   <li>the XML <B>ASSEMBLAGE</B> {@link org.objectweb.jonathan.tools.compilers.AssemblageCompiler tag compiler},
 *   <li>the XML <B>INCLUDE</B> {@link org.objectweb.jonathan.tools.compilers.IncludeCompiler tag compiler}
 *   <li>the XML <B>PROPERTY</B> {@link org.objectweb.jonathan.tools.compilers.PropertyCompiler tag compiler}
 *   <li>the XML <B>IMPLICIT_FACTORY</B> {@link org.objectweb.jonathan.tools.compilers.ImplicitFactoryCompiler tag compiler}
 *   <li>the XML <B>SEQUENCE</B> {@link org.objectweb.jonathan.tools.compilers.SequenceCompiler tag compiler}
 * </ul>
 * A custom tag compiler has to implement the {@link
 * org.objectweb.jonathan.tools.compilers.TagCompiler TagCompiler} interface
 * and to be provided as a line like the next one in the Kcf2java
 * properties file:<BR>
 * CUSTOM_XML_TAG_TO_COMPILE my.custom.tag.Compiler
 */
public class Kcf2java {
   static boolean compile = true;
   static boolean verbose = false;
   static String src_dir = null;
   static String properties_file_name = null;
   static String class_dir = null;
   static String configuration_file_name;
   static String java_file;
   static String java_ext = ".java";
   static String class_ext = ".class";
   static String class_name;
   static String package_name;
   static String scoped_name;
   
   static int configuration_number = 0;

   static Properties compilers_specs = new Properties(System.getProperties());
   static {
      // setting defaults
      compilers_specs.put("ALIAS", "org.objectweb.jonathan.tools.compilers.AliasCompiler");
      compilers_specs.put("ATOM", "org.objectweb.jonathan.tools.compilers.AtomCompiler");
      compilers_specs.put("ASSEMBLAGE", "org.objectweb.jonathan.tools.compilers.AssemblageCompiler");
      compilers_specs.put("INCLUDE", "org.objectweb.jonathan.tools.compilers.IncludeCompiler");
      compilers_specs.put("PROPERTY", "org.objectweb.jonathan.tools.compilers.PropertyCompiler");
      compilers_specs.put("IMPLICIT_FACTORY", "org.objectweb.jonathan.tools.compilers.ImplicitFactoryCompiler");
      compilers_specs.put("SEQUENCE", "org.objectweb.jonathan.tools.compilers.SequenceCompiler");
   }
   


   static Hashtable configuration_names = new Hashtable();

   static Hashtable tag_compilers = new Hashtable();

   /** 
    * The main method.
    * @param args see the description of the {@link Kcf2java Kcf2java} class. 
    */
   public static void main (String[] args){
      parseArgs(args);
      InputStream $is = null;
      try {

         if (properties_file_name != null) {
            compilers_specs.load(Kcf2java.class.getClassLoader().getResourceAsStream(properties_file_name));
         }

         try{
            $is = new FileInputStream(configuration_file_name);
            compile($is,src_dir,false,null,verbose,compile);
         } catch (IOException e){
            e.printStackTrace();
            throw e;
         }

      } catch (Exception e) {
         e.printStackTrace();
         System.exit(1);
      } finally {
         try {
            if ($is != null) {
               $is.close();
            }
         } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
         }
      }
   }


   /** 
    * Compilation method to be used when Kcf2java is embedded in some application.
    * @param _is an InputStream to the configuration to compile;
    * @param _src_dir the directory where the source files should be generated;
    * @param _add_package_path_to_src_dir true if Kcf2java should generate
    *        the file taking into account the package name of the class to
    *        generate, false if the file should be generated directly in <quote>_src_dir</quote> 
    * @param _custom_properties the properties representing custom tag decoders;
    * @param _verbose indicates whether error messages and warnings should be
    *                 output to stderr
    * @param _compile indicates whether the generated Java file should be compiled.
    * @exception JonathanException if something goes wrong.
    */
   public static void compile(InputStream _is,
                              String _src_dir,
                              boolean _add_package_path_to_src_dir,
                              Properties _custom_properties,
                              boolean _verbose,
                              boolean _compile)
      throws JonathanException {
      
      verbose = _verbose;

      File $target_java_file = null; // to be deleted if compilation fails

      try {

         if (_custom_properties != null) {
            Enumeration $names = _custom_properties.propertyNames();
            while ($names.hasMoreElements()) {
               String $name = (String) $names.nextElement();
               String $value = (String) _custom_properties.get($name);
               compilers_specs.setProperty($name,$value);
            }
         }

         XMLElement $xmlElement = new XMLElement();
      
         try{
            $xmlElement.parseFromReader(new BufferedReader(new InputStreamReader(_is)));

         } catch (IOException e){
            e.printStackTrace();
            return;
         }
      
         String $currentTag = $xmlElement.getName();
         if (!$currentTag.equals("CONFIGURATION")){
            throw new JonathanException("not a Jonathan Configuration File !");
         }

         String $kernelClass = $xmlElement.getStringAttribute("BOOTSTRAP");
         if ($kernelClass==null || $kernelClass.equals("")){
            throw new JonathanException("Error: No class name specified for the bootstrap Kernel.\n"
               +"The \"bootstrap\" attribute of the first tag is missing."
               +"\nIt specifies the class name of the bootstrap Kernel."
               +"\nThe first tag of the file should be something like :"
               +"\n<CONFIGURATION bootstrap=\"a.package.name.BootstrapKernelClass\""
               +"\n<              generated=\"a.package.name.GeneratedKernelClass\">");
         }

         scoped_name = $xmlElement.getStringAttribute("GENERATED");
         if (scoped_name == null || scoped_name.equals("")) {
            throw new JonathanException("Error: No class name specified for the generated Kernel."
               +"\nThe \"generated\" attribute of the first tag is missing." 
               +"\nIt specifies the class name of the generated Kernel."
               +"\nThe first tag of the file should be something like :"
               +"\n<CONFIGURATION bootstrap=\"a.package.name.BootstrapKernelClass\""
               +"\n<              generated=\"a.package.name.GeneratedKernelClass\">");
         } else {
            int last_dot = scoped_name.lastIndexOf('.');
            if (last_dot < 0) {
               class_name = scoped_name;
            } else {
               class_name = scoped_name.substring(last_dot + 1);
               package_name = scoped_name.substring(0,last_dot);
            }
         }

         scoped_name =  scoped_name.replace('.',File.separatorChar);
         String $src_dir = _src_dir;
         if ($src_dir == null) {
            $src_dir = "";
         }
         if (!$src_dir.endsWith(File.separator)) {
            $src_dir = $src_dir + File.separator;
         }
         if (_add_package_path_to_src_dir && package_name != null) {
             $src_dir = $src_dir + package_name.replace('.',File.separatorChar)
                + File.separator;
         }
         java_file = $src_dir + class_name + java_ext;

         $target_java_file = new File(java_file);
         //some sanity checks
         if ($target_java_file.exists() && $target_java_file.isDirectory()) {
            //unable to create taget java file
            throw new JonathanException("Kcf to java compilation: unable to compile; " +
                                        "target file is a directory!");
         }
         File $target_file_parent = $target_java_file.getParentFile();
         if (!$target_file_parent.exists()) {
            //parent directory has to be created first
            $target_file_parent.mkdirs();
         }
         if (!$target_java_file.exists()) {
            //parent directory has to be created first
            $target_java_file.createNewFile();
         }

         FileWriter writer = new FileWriter($target_java_file);

         writer.write("// Generated File : do not edit\n");
         if (package_name != null) {
            writer.write("package " + package_name + ";\n\n");
         }
         writer.write("import org.objectweb.jonathan.apis.kernel.Context;\n");
         writer.write("import org.objectweb.jonathan.apis.kernel.Component;\n");
         writer.write("import org.objectweb.jonathan.apis.kernel.JonathanException;\n");
         writer.write("import org.objectweb.jonathan.apis.kernel.InternalException;\n");
         writer.write("import org.objectweb.jonathan.libs.kernel.JConfigurationFactory;\n");
         writer.write("import org.objectweb.jonathan.libs.kernel.JComponent;\n\n");
         if (configuration_file_name != null) {
             writer.write("/**\n");
             writer.write(" * This class provides a method to compute the bootstrap configuration,\n");
             writer.write(" * generated from the file " + configuration_file_name.replace('\\', '/') + ".\n");
             writer.write(" */\n");
         }
         writer.write("public final class " + class_name + " {\n");
         writer.write("   static final JConfigurationFactory context_factory = new JConfigurationFactory();\n");
         if (configuration_file_name != null) {
             writer.write("  /**\n");
             writer.write("    * This configuration represents the configuration specified in " +
                          configuration_file_name.replace('\\', '/') + ".\n");
             writer.write("    * It is used to generate the bootstrap configuration.\n");
             writer.write("    */\n");
         }
         writer.write("   private static final Context initial_configuration = context_factory.newContext();\n");
         writer.write("   static {\n");
         writer.write("      try{\n");
         String prop;

         Enumeration $elements = $xmlElement.enumerateChildren();
         //writeElements("initial_configuration",$elements,writer,"",null);
         writeElements("initial_configuration",$elements,writer,"");
         
         // HACK!! allows the "try .. catch" to be compiled if there is no component
         // in the root configuration
         writer.write("         if (1 == 0) throw new JonathanException();\n");

         writer.write("      } catch(JonathanException e){\n");
         writer.write("         //Cannot happen. \n");
         writer.write("         e.printStackTrace(System.err);\n");
         writer.write("         throw new InternalException(\"Unexpected exception.\");\n");
         writer.write("      }\n");
         writer.write("   }\n");
         writer.write("    \n");
         writer.write("   /**\n");
         writer.write("    * Returns the effective bootstrap configuration.\n");
         writer.write("    * @param c the class of the caller of this method.\n");
         writer.write("    * @return the bootstrap configuration.\n");
         writer.write("    */\n");
         writer.write("   public static Context newConfiguration(Class c){\n");
         writer.write("      return "+$kernelClass+".newConfiguration(initial_configuration,c);\n");
         writer.write("   }\n");
         writer.write("}");
         writer.close();

         if (_compile) {
            Process process = null;
            if (class_dir == null) {
               String[] command = { "javac", java_file };
               message("Compiling " + java_file +
                       " with javac");
               process = Runtime.getRuntime().exec(command);
            } else if (class_dir.endsWith(".jar") || class_dir.endsWith(".zip")) {
               String[] command = { "javac", "-d" , ".", java_file };
               message("Compiling " + java_file + " with javac -d .");
               process = Runtime.getRuntime().exec(command);
            } else {
               String[] command = { "javac", "-d" , class_dir, java_file };
               message("Compiling " + java_file + " with javac -d " + class_dir);
               process = Runtime.getRuntime().exec(command);
            } 
            process.waitFor();
            int exit = process.exitValue();
            if (exit == 0) {
               message("Kernel successfully compiled.");
            } else {
               throw new JonathanException("Compilation exited abnormally with code " + exit);
            }
            if (class_dir != null && class_dir.endsWith(".jar")) {
               String class_file =  scoped_name + class_ext;
               String[] command = { "jar", "uf", class_dir, class_file};
               message("Adding " + class_file + " to " + class_dir);
               process = Runtime.getRuntime().exec(command);
               process.waitFor();
               exit = process.exitValue();
               if (exit == 0) {
                  message("Kernel successfully added.");
               } else {
                  throw new JonathanException("Abnormal exit with code " + exit);
               }
            }
         }

      } catch (Exception e) {
         try {
            $target_java_file.delete();
         } catch (Exception _del_ex) {
            throw new JonathanException(_del_ex + " while " + e);
         }
         throw new JonathanException(e);

      } finally {
         try {
            if (_is != null) {
               _is.close();
            }
         } catch (Exception e) {
            throw new JonathanException(e);
         }
      }
   }
   
   
   static void parseArgs(String[] args) {
      int len = args.length;
      for (int i = 0; i < len; i++) {
         if (args[i].equals("-d")) {
            if (class_dir == null) {
               class_dir = args[++i];
            } else {
               throw new RuntimeException("Class path is already defined.");
            }
         } else if (args[i].equals("-noc")) {
            compile = false;
         } else if (args[i].equals("-s")) {
            if (src_dir == null) {
               src_dir = args[++i];
            } else {
               throw new RuntimeException("Source path is already defined.");
            }
         } else if (args[i].equals("-p")) {
            if (properties_file_name == null) {
               properties_file_name = args[++i];
            } else {
               throw new RuntimeException("Properties file name is already defined.");
            }
         } else if (args[i].equals("-verbose")) {
            verbose = true;
         } else {
            if (configuration_file_name == null) {
               configuration_file_name = args[i];
            } else {
               throw new RuntimeException("Kernel file is already defined.");
            }
         }
      }
   }
   
   /** 
    * Normalizes the provided string, by trimming and handling backslashes
    * @param str the string to normalize
    * @return the normalized string
    */
   public static String normalize(String str) {
      int prev = 0, index;
      String ret = "";
      while ((index = str.indexOf('\\',prev)) >= 0) {
         ret = ret + str.substring(prev,index) + "\\\\";
         prev = index + 1;
      }
      if (prev == 0) {
         return str.trim();
      } else {
         return ret.trim();
      }
   }

   /** 
    * Outputs a message (the provided string) on stderr, when the compiler
    * is started with the -verbose option.
    * @param str a message to output.
    */
   static void message(String str) {
      if (verbose) {
         System.err.println(str);
      }
   }
   
   private static void writeElements(String _configuration_var_name,
                                     Enumeration _enum,
                                     Writer _writer,
                                     String _configuration_name)
      throws IOException {
      XMLElement $current_xml_element = null;
      String $current_tag = null;
      String $element_name = null;
      
      while (_enum.hasMoreElements()) {
         $current_xml_element = (XMLElement) _enum.nextElement();
         $current_tag = $current_xml_element.getName();
         if ($current_tag.equals("DOCUMENTATION")) {
            // skip
         } else if (! $current_tag.equals("ELEM")) {
            throw new IOException("At line " + $current_xml_element.getLineNr()
                                  + ": found tag \"" + $current_tag + 
                                  "\" whereas \"ELEM\" was expected.");
         } else if ($current_xml_element.countChildren() == 0) { //unset component
            //skip
         } else {
            $element_name = normalize($current_xml_element.getStringAttribute("NAME"));
            XMLElement $current_component = (XMLElement) $current_xml_element.getChildren().elementAt(0);
            _writer.write("         "+_configuration_var_name+".addElement(\"" + 
                          normalize($element_name) + "\"," + 
                          writeComponent($current_component,$element_name,_configuration_var_name,
                                         _writer,_configuration_name) +
                          ",'/');\n");
         }
      }
   }

   /** 
    * Compiles the provided component using the provided writer. 
    * @param _current_component the component to compile
    * @param _element_name its name in the current configuration
    * @param _configuration_var_name the variable name of the current configuration (may be null)
    * @param _writer the writer to which the compiled version is output
    * @param _configuration_name the absolute name of the current configuration
    * @return the name of the variable used in the Java compiled file to represent the compiled component
    * @exception IOException if something goes wrong.
    */
   static public String writeComponent(XMLElement _current_component,
                                       String _element_name,
                                       String _configuration_var_name,
                                       Writer _writer,
                                       String _configuration_name)
      throws IOException {

      String $current_tag = _current_component.getName();
      
      if ($current_tag.equals("CONFIGURATION")){
         String $var_name = "CONFIGURATION" + (configuration_number++);
         String $configuration_element_name = normalize(_element_name);
         String $scope_name = _current_component.getStringAttribute("SCOPE");
         String $new_configuration_name = _configuration_name + "/" + $configuration_element_name;
         if ($scope_name != null) {
            $scope_name = normalize($scope_name);
            if (! $scope_name.startsWith("/")) {
               $scope_name = _configuration_name + "/" + $scope_name;
            }
         } 

         if (_configuration_var_name != null) {
            _writer.write("         Context "+ $var_name + 
                          " = " + _configuration_var_name + ".addOrGetContext(\"" + 
                          $configuration_element_name + "\",'/');\n");
         } else {
            _writer.write("         Context "+ $var_name + 
                          " = context_factory.newContext();\n");
         }            
            
         String scope = $scope_name != null ? (String) configuration_names.get($scope_name) : null;
         if (scope != null) {
            _writer.write("         " + $var_name + 
                          ".setScope(" + scope + ");\n");
         }
         configuration_names.put($new_configuration_name,$var_name);
         
         Enumeration $elements = _current_component.enumerateChildren();            
         writeElements($var_name,$elements,_writer,
                       $new_configuration_name);
         
         return $var_name;
      } else {
         TagCompiler $compiler = getTagCompiler($current_tag);
         if ($compiler == null) {
            throw new IOException("No compiler found for tag " + $current_tag + " found at line " +
                                  _current_component.getLineNr());
         }
         return $compiler.writeComponent(_current_component,_element_name,_configuration_var_name,
                                         _writer,_configuration_name);
      }
   }

   static TagCompiler getTagCompiler(String _tag) throws IOException {
      Object $component = tag_compilers.get(_tag);
      if ($component instanceof TagCompiler) {
         return (TagCompiler) $component;
      } else {
         String $tag_class = compilers_specs.getProperty(_tag);
         if ($tag_class != null) {
            Class $tag_compiler_class = null;
            try {
               $tag_compiler_class = Class.forName($tag_class);
            } catch (Exception e) {
               throw new IOException("Can't load tag compiler class " + $tag_class + ": " + e);
            }
            try {
               $component = $tag_compiler_class.newInstance();
            } catch (Exception e) {
               throw new IOException("Can't instantiate tag compiler of class " + $tag_class + ": " + e);
            }
            if ($component instanceof TagCompiler) {
               tag_compilers.put(_tag,$component);
               return (TagCompiler) $component;
            } else {
               throw new IOException("Bad tag compiler class: " + $tag_class + 
                                     " is not a subtype of TagCompiler." );
            }
         } else {
            throw new IOException("No compiler registered for tag " + _tag + ".");
         }
      }
   }
}


   
   

